import { Inject, Injectable } from '@angular/core';
import { NetworkService } from '../network/network.service';
import { ApiPromise } from '@polkadot/api';
import { Account, AccountUtils, DerivationAccount, LoginOptions, SelectAccountOptions } from './account.model';
import { keyring } from '@polkadot/ui-keyring';
import { environment } from '@environments/environment';
import { KeyringStorage } from '@app/shared/services/storage/keyring-storage';
import { base58Encode, cryptoWaitReady, encodeAddress, mnemonicGenerate } from '@polkadot/util-crypto';
import {
  firstArrayValue,
  isEmptyArray,
  isNil,
  isNilOrBlank,
  isNilOrNaN,
  isNotEmptyArray,
  isNotNil,
  isNotNilOrBlank,
  isNotNilOrNaN,
  sleep,
  toNumber,
} from '@app/shared/functions';
import { APP_STORAGE, IStorage } from '@app/shared/services/storage/storage.utils';
import {
  combineLatest,
  debounceTime,
  firstValueFrom,
  from,
  map,
  mergeMap,
  Observable,
  startWith,
  Subject,
  Subscription,
  switchMap,
  takeUntil,
  tap,
  timer,
} from 'rxjs';
import { Currency } from '@app/currency/currency.model';
import { SettingsService } from '@app/settings/settings.service';
import { scryptEncode } from '@polkadot/util-crypto/scrypt/encode';
import { u8aToHex } from '@polkadot/util';
import { formatAddress, formatPubkey } from '@app/shared/currencies';
import { RxStartableService } from '@app/shared/services/rx-startable-service.class';
import { RxStateProperty, RxStateSelect } from '@app/shared/decorator/state.decorator';
import { ED25519_SEED_LENGTH, SCRYPT_PARAMS } from '@app/account/crypto.utils';
import { KeyringPair } from '@polkadot/keyring/types';
import { IndexerService } from '@app/network/indexer/indexer.service';
import { AppEvent } from '@app/shared/types';
import { APP_AUTH_CONTROLLER, AuthData, IAuthController } from '@app/account/auth/auth.model';
import { ExtrinsicError, ExtrinsicUtils } from '@app/shared/substrate/extrinsic.utils';
import { IdentityStatusEnum } from '@app/network/indexer/indexer-types.generated';
import { PodService } from '@app/network/pod/pod.service';
import { FetchPolicy } from '@apollo/client';
import { TranslateService } from '@ngx-translate/core';
import { WotService } from '@app/wot/wot.service';
import { CertificationSearchFilter } from '@app/certification/history/cert-history.model';
import { ISubmittableResult } from '@polkadot/types/types';
import { BiometricAuth } from '@aparajita/capacitor-biometric-auth';
import { AlertController, ModalController } from '@ionic/angular';
import { ValidationModal } from '@app/transfer/send/transfer-validation/validation.modal';

// kind of certify action
enum CertType {
  // first certification is actually an identity creation
  IdentityCreation,
  // normal cert creation
  CertCreation,
  // cert renewal
  CertRenewal,
}

export interface LoadAccountDataOptions {
  reload?: boolean;
  withTx?: boolean;
  withCert?: boolean;
  withBalance?: boolean;
  withMembership?: boolean;
  emitEvent?: boolean;
  withProfile?: boolean;
  fetchPolicy?: FetchPolicy;
}

export interface WatchAccountDataOptions extends LoadAccountDataOptions {}

export interface AccountsState {
  accounts: Account[];
  password?: string;
}

@Injectable({ providedIn: 'root' })
export class AccountsService extends RxStartableService<AccountsState> {
  private _listenChangesSubscription: Subscription = null;
  private _store = new KeyringStorage(this.storage);
  private _passwordTimer: Subscription;
  private _keyringInitialized = false;

  get api(): ApiPromise {
    return this.network.api;
  }

  @RxStateProperty() accounts: Account[];
  @RxStateSelect() accounts$: Observable<Account[]>;

  @RxStateProperty() private _password: string;

  get isLogin(): boolean {
    return this.started && isNotEmptyArray(this.accounts);
  }

  constructor(
    protected network: NetworkService,
    protected indexer: IndexerService,
    protected wot: WotService,
    protected pod: PodService,
    protected settings: SettingsService,
    protected translate: TranslateService,
    protected modalCtrl: ModalController,
    protected alertCtrl: AlertController,
    @Inject(APP_STORAGE) protected storage: IStorage,
    @Inject(APP_AUTH_CONTROLLER) protected authController: IAuthController
  ) {
    super(network, {
      name: 'account-service',
    });
    this.start();

    this.hold(this.accounts$, (accounts) => this.listenAccountsData(accounts));
  }

  protected async ngOnStart(): Promise<AccountsState> {
    // Wait crypto to be loaded by browser
    await cryptoWaitReady();

    const currency = this.network.currency;

    // Configure keyring
    keyring.setDevMode(!environment.production);
    keyring.setSS58Format(currency.ss58Format || 42 /* = dev format */);

    // Restoring accounts
    const accounts = await this.restoreAccounts(currency);

    return <AccountsState>{
      accounts,
    };
  }

  listenAccountsData(accounts: Account[]) {
    console.debug(`${this._logPrefix}Listening accounts changes...`);
    const accountsObservables = accounts.map((account) =>
      this.watchData(account).pipe(
        tap((account) => {
          account.changes?.next();
        })
      )
    );
    this._listenChangesSubscription?.unsubscribe();
    this._listenChangesSubscription = combineLatest(accountsObservables).subscribe();
  }

  protected watchData(account: Account): Observable<Account> {
    return new Observable<Account>((observer) => {
      let previousFree: number = account.data.free;
      let teardown: VoidFunction;

      this.network.api.query.system
        .account(account.address, ({ data }) => {
          // Get balance delta
          const change = toNumber(data.free) - previousFree;

          // Update the balance if changed
          if (change !== 0) {
            account.data = {
              ...account.data,
              ...data.toJSON(),
            };
            console.debug(`${this._logPrefix}New account balance for ${account.address}...`);
            previousFree = toNumber(data.free);
            observer.next(account);
          }
        })
        .then((res) => {
          teardown = res;
        });

      return () => {
        if (teardown) teardown();
      };
    });
  }

  protected async ngOnStop(): Promise<void> {
    this._keyringInitialized = false;
    return super.ngOnStop();
  }

  selectAccount(opts?: SelectAccountOptions): Promise<Account> {
    return this.authController.selectAccount(opts);
  }

  async createNew(opts?: { redirectToWalletPage?: boolean }): Promise<Account> {
    const data = await this.authController.createNew(opts);

    if (!data?.address) return null; // User cancelled

    // Make account exists
    if (!(await this.isAvailable(data.address))) {
      throw { message: 'ERROR.UNKNOWN_WALLET_ID1' };
    }

    return data;
  }

  async login(event?: AppEvent, opts?: LoginOptions): Promise<Account> {
    const data = await this.authController.login(event, opts);

    if (!data?.address) return null; // User cancelled

    // Make account exists
    if (!(await this.isAvailable(data.address))) {
      throw { message: 'ERROR.UNKNOWN_WALLET_ID1' };
    }

    return data;
  }

  async restoreAccounts(currency?: Currency) {
    // load all available addresses and accounts
    const now = Date.now();
    console.info('[account-service] Loading all accounts...');

    // Prepare an observable, to known when keyring.loadAll() will be ready
    const accounts$ = keyring.accounts.subject.pipe(
      debounceTime(250),
      map(() => keyring.getAccounts())
    );

    keyring.loadAll({
      store: this._store,
      ss58Format: currency?.ss58Format,
      genesisHash: currency?.genesis,
      isDevelopment: !environment.production,
    });

    const keyringAddresses = await firstValueFrom(accounts$);
    this._keyringInitialized = true;
    if (isEmptyArray(keyringAddresses)) {
      console.info('[account-service] Loading all accounts [OK] No account found');
    } else {
      const accounts = keyringAddresses.map((ka) => {
        return <Account>{
          address: ka.address,
          publicKey: ka.publicKey,
          meta: {
            ...ka.meta,
            self: true,
          },
          changes: new Subject<void>(),
        };
      });

      // Add Dev account
      // if (!environment.production) {
      //   const devAccount = await this.addDevAccount(accounts);
      //   if (devAccount) accounts.push(devAccount);
      // }

      // Load account's data
      try {
        await Promise.all(accounts.map((a) => this.loadData(a, { withMembership: true, withProfile: true })));

        // DEBUG
        console.info(this._logPrefix + `Loading accounts [OK] ${accounts.length} accounts loaded in ${Date.now() - now}ms`);
        accounts.forEach((a) => {
          console.info(` - ${AccountUtils.getDisplayName(a)} {free: ${a.data?.free / 100}, reserved: ${a.data?.reserved / 100}}`);
        });

        return accounts;
      } catch (err) {
        console.error('Error while loading accounts', err);
        return [];
      }
    }
  }

  /**
   * Add test account --- DEV only
   */
  protected async addDevAccount(accounts: Account[]): Promise<Account> {
    const data = { ...environment.dev?.auth, changes: new Subject<void>() };

    // Set password to AAAAA (or those defined in environment)
    this._password = data?.password || '12345';

    if (!data || (!data.v1 && !data.v2)) return; // Skip if no dev account defined
    data.meta = {
      isTesting: true,
      default: true,
      ...data.meta,
      self: true,
      isMember: true,
    };

    const { pair, account } = await this.createAccount(data);

    const index = (accounts || []).findIndex((a) => a.address === pair.address) || -1;
    if (index !== -1) {
      accounts[index] = {
        ...accounts[index],
        ...data,
      };
      // Update meta
      keyring.saveAccountMeta(pair, pair.meta);
      return undefined;
    }

    // Add new account
    else {
      keyring.saveAccount(pair, this._password);
      return account;
    }
  }

  async addAccount(auth: AuthData): Promise<Account> {
    if (!auth) return;

    if (!this.started) await this.ready();

    const { pair, account } = await this.createAccount(auth);

    const isAuth = await this.auth();
    if (!isAuth) return; // Skip is cannot write to keyring

    await this.storage.set('validationCode', this._password);
    // Save account into keyring
    keyring.saveAccount(pair, this._password);

    // Save account into service's list
    await this.saveAccount(account);

    return account;
  }

  async auth(): Promise<boolean> {
    if (isNotNilOrBlank(this._password)) {
      console.debug(`${this._logPrefix}Already authenticated. Skip`);
      return true; // ok
    }

    // const hasPassword = this.settings.get('hasPassword') || false;  //TODO: TO BE VALIDED
    // if (!hasPassword) {
    //   console.error(`${this._logPrefix}No password defined. Cannot auth`);
    //   return false; // KO
    // }

    console.debug(`${this._logPrefix}Not auth: opening unlock modal...`);

    const data = await this.authController.unlock();

    // User cancelled
    if (isNilOrBlank(data)) {
      console.debug(`${this._logPrefix}Not auth: cancelled`);
      return false;
    }

    this._password = data;

    // Un auth after a delay
    this._passwordTimer?.unsubscribe();
    const resetDelay = Math.max(this.settings.get('unAuthDelayMs') || 0, 5000); // 5s min
    // eslint-disable-next-line @rx-angular/no-zone-critical-rxjs-creation-apis
    this._passwordTimer = timer(resetDelay).subscribe(() => {
      if (isNotNil(this._password)) {
        this._password = null;

        // Lock all pairs
        (this.accounts || [])
          .map((a) => keyring.getPair(a.address))
          .filter((pair) => pair.isLocked)
          .forEach((pair) => pair.lock());
      }
      this._passwordTimer?.unsubscribe();
      this._passwordTimer = null;
    });
    return true;
  }

  async generateMnemonic(numWords: 12 | 15 | 18 | 21 | 24 = 12) {
    if (!this._keyringInitialized) await this.ready();

    // generate a random mnemonic
    return mnemonicGenerate(numWords);
  }

  generateAddress(mnemonicWithDerivation: string): string {
    try {
      const pair = keyring.createFromUri(mnemonicWithDerivation, null, 'sr25519');
      return encodeAddress(pair.address);
    } catch (error) {
      console.error('Error generating address from mnemonic:', error);
      return '';
    }
  }

  async createPair(data: AuthData): Promise<KeyringPair> {
    if (!this._keyringInitialized) await this.ready();

    if (data.v1) {
      const passwordU8a = Uint8Array.from(data.v1.password.split('').map((x) => x.charCodeAt(0)));
      const saltU8a = Uint8Array.from(data.v1.salt.split('').map((x) => x.charCodeAt(0)));
      const result = scryptEncode(passwordU8a, saltU8a, data.v1.scryptParams || SCRYPT_PARAMS.DEFAULT);
      const seedHex = u8aToHex(result.password.slice(0, ED25519_SEED_LENGTH));
      return keyring.createFromUri(seedHex, data.meta || {}, 'ed25519');
    } else if (data.v2) {
      return keyring.createFromUri(data.v2.mnemonic, data.meta, 'sr25519');
    }
  }

  async createAccount(data: AuthData): Promise<{ account: Account; pair: KeyringPair }> {
    const pair = await this.createPair(data);
    const publicKeyV1 = pair.type === 'ed25519' ? base58Encode(pair.publicKey) : undefined;

    const account: Account = {
      address: pair.address,
      publicKey: pair.publicKey,
      meta: {
        ...pair.meta,
        name: data.meta?.name || (publicKeyV1 ? formatPubkey(publicKeyV1) : formatAddress(pair.address)),
        publicKeyV1: publicKeyV1,
        genesisHash: this.network.currency?.genesis,
        ...data.meta,
      },
    };

    pair.setMeta(account.meta);

    return { pair, account };
  }

  async saveAccount(account: Account): Promise<Account> {
    const accounts = this.accounts || [];

    // Define as default
    if (account.meta?.default || accounts.length === 1) this.setDefaultAccount(account, accounts);

    await this.loadData(account);

    // Append to accounts
    if (this.started) {
      const alreadyExists = await this.isAvailable(account.address);
      if (!alreadyExists) {
        this.accounts = [...(this.accounts || []), account];
      }
    }

    return account;
  }

  setDefaultAccount(account: Account, accounts?: Account[]) {
    account.meta = {
      ...account.meta,
      default: true,
    };
    // Set other as NOT default
    accounts = accounts || this.accounts || [];
    accounts
      .filter((a) => a.address !== account.address && a.meta?.default)
      .forEach((a) => {
        a.meta = {
          ...a.meta,
          default: false,
        };
      });
  }

  watchAll(opts?: { positiveBalanceFirst?: boolean; minBalance?: number; isMember?: boolean }): Observable<Account[]> {
    if (!this.started) {
      return from(this.ready()).pipe(switchMap(() => this.watchAll(opts)));
    }

    return this.select('accounts').pipe(
      map((accounts) => {
        // Filter is member
        if (isNotNil(opts?.isMember)) {
          accounts = accounts.filter((a) => (a.meta.isMember || false) === opts.isMember);
        }

        // Filter min balance
        if (isNotNil(opts?.minBalance)) {
          accounts = accounts.filter((a) => AccountUtils.getBalance(a) >= opts.minBalance);
        }

        // Sort with a balance first
        if (opts?.positiveBalanceFirst) {
          accounts.sort((a1, a2) => {
            const b1 = AccountUtils.getBalance(a1);
            const b2 = AccountUtils.getBalance(a2);
            return b1 === b2 ? 0 : b1 < b2 ? 0 : -1;
          });
        }
        return accounts;
      })
    );
  }

  async getAll(): Promise<Account[]> {
    if (!this.started) await this.ready();

    return this.accounts || [];
  }

  async getDefault(opts?: LoadAccountDataOptions): Promise<Account> {
    if (!this.started) await this.ready();

    const accounts = this.accounts || [];
    let account = accounts.find((a) => a.meta?.default);
    if (!account) {
      if (accounts.length) {
        account = accounts[0];
        this.setDefaultAccount(account);
      } else {
        throw { message: 'ERROR.UNKNOWN_WALLET_ID1' };
      }
    }

    // Load data
    return await this.loadData(account, opts);
  }

  async getByName(name: string, opts?: LoadAccountDataOptions): Promise<Account> {
    if (!this.started) await this.ready();

    const accounts = this.accounts || [];

    const account = accounts.find((a) => a.meta?.name === name);
    if (!account) {
      throw { message: 'ERROR.UNKNOWN_WALLET_ID2' };
    }

    // Load
    return await this.loadData(account, opts);
  }

  async isAvailable(address: string, accounts?: Account[]): Promise<boolean> {
    if (!accounts && !this.started) await this.ready();
    return (accounts || this.accounts || []).some((a) => a.address === address);
  }

  isAvailableSync(address: string, accounts?: Account[]): boolean {
    return (accounts || this.accounts || []).some((a) => a.address === address);
  }

  async getByAddress(address: string, opts?: LoadAccountDataOptions): Promise<Account> {
    if (!this.started) await this.ready();

    const accounts = this.accounts || [];
    const account = accounts.find((a) => a.address === address);
    if (!account) {
      throw { message: 'ERROR.UNKNOWN_WALLET_ID3' };
    }

    // Load data
    return await this.loadData(account, opts);
  }

  /*private cacheByAddress: {
    [address: string]: {
      time: number;
      options: WatchAccountDataOptions;
      account$: Observable<Account>;
    };
  } = {};*/

  watchByAddress(address: string, opts?: WatchAccountDataOptions): Observable<Account> {
    // Wait start if need, then loop
    if (!this.started) return from(this.ready()).pipe(switchMap(() => this.watchByAddress(address, opts)));

    if (this.isAvailableSync(address)) {
      const stopWatching = new Subject<void>();
      return this.accounts$.pipe(
        map((accounts) => accounts?.find((a) => a.address === address)),
        tap((account) => account || stopWatching.next()),
        mergeMap(async (account) => {
          return this.loadData(account, {
            ...opts,
            withMembership: false,
            fetchPolicy: 'cache-first',
          });
        }),
        mergeMap((account) =>
          account.changes.pipe(
            startWith(),
            map(() => account)
          )
        ),
        takeUntil(stopWatching)
      );
    }

    // TODO use cache
    // Exists in cache
    //if (this.cacheByAddress[address]) {
    //}

    return this.indexer.wotSearch({ address }, { first: 1, fetchPolicy: 'cache-first' }).pipe(
      map(({ data }) => firstArrayValue(data) || { address, meta: { name: formatAddress(address) } }),
      mergeMap(async (account) =>
        this.loadData(account, {
          ...opts,
          withMembership: false,
          fetchPolicy: 'cache-first',
        })
      )
    );
  }

  /**
   * Transfer some money to an address
   * @param from
   * @param to
   * @param amount the TX amount, using decimals
   * @param comment
   * @param fee the TX fee, using decimals
   */
  async transfer(from: Partial<Account>, to: Partial<Account>, amount: number, comment?: string, fee?: number): Promise<string> {
    if (!from || !to) throw new Error("Missing argument 'from' or 'to' !");
    const currency = this.network.currency;

    // Check currency
    if (!currency) throw new Error('ERROR.CHECK_NETWORK_CONNECTION');
    const powBase = Math.pow(10, currency.decimals || 0);

    // Check amount
    if (isNilOrNaN(amount)) {
      throw new Error('ERROR.AMOUNT_REQUIRED');
    }
    if (amount < 0) {
      throw new Error('ERROR.AMOUNT_NEGATIVE');
    }

    // Remove decimals, in amount and fee
    amount = amount * powBase;
    if (fee) fee = fee * powBase;

    // Check fee validity
    fee = fee || currency.fees?.tx || 0;
    if (fee < 0) {
      throw new Error('ERROR.FEE_NEGATIVE');
    }

    // Check issuer != recipient
    if (from.address === to.address) {
      throw new Error('ERROR.SAME_TX_RECIPIENT');
    }

    // Get issuer account
    const issuerAccount = await this.getByAddress(from.address);

    // Check enough credit
    if (amount + fee > issuerAccount.data.free) {
      throw new Error('ERROR.NOT_ENOUGH_CREDIT');
    }

    console.info(
      `[account-service] Sending ${amount / powBase} ${currency.symbol} (fee: ${fee / powBase}):\nfrom: ${from.address}\nto ${to.address}`
    );

    // Get pair, and unlock it
    const issuerPair = keyring.getPair(issuerAccount.address);
    if (issuerPair.isLocked) {
      console.debug(`[account-service] Unlocking address ${from.address} ...`);
      const isAuth = await this.auth();
      if (!isAuth) throw new Error('ERROR.AUTH_REQUIRED');
      issuerPair.unlock(this._password);
    }

    try {
      let result: ISubmittableResult;
      // Transfer with comment
      if (isNotNilOrBlank(comment)) {
        const txTransfer = this.api.tx.balances.transferKeepAlive(to.address, amount);
        const txComment = this.api.tx.system.remarkWithEvent(comment ?? '');
        result = await ExtrinsicUtils.submit(this.api.tx.utility.batch([txTransfer, txComment]), issuerPair);
      } else {
        // Transfer without comment
        result = await ExtrinsicUtils.submit(this.api.tx.balances.transferKeepAlive(to.address, amount), issuerPair);
      }
      const { status, events } = result;
      console.info(`${this._logPrefix}Extrinsic status`, status.toHuman());
      console.info(`${this._logPrefix}Completed at block hash #${status.hash.toHuman()}`);

      if (this._debug) console.debug(`${this._logPrefix}Block events:`, JSON.stringify(events));

      // List of outdated accounts
      const outdatedAccounts = [issuerAccount];

      // Add receiver to outdated account
      if (await this.isAvailable(to.address)) {
        const toAccount = await this.getByAddress(to.address);
        outdatedAccounts.push(toAccount);
      }

      await sleep(200); // Wait 200ms

      await this.refreshData(outdatedAccounts, { reload: true });

      return status.hash.toString();
    } catch (err) {
      console.error(err);
      throw new Error('ERROR.SEND_TX_FAILED');
    }
  }

  /**
   *
   * @param from
   * @param to
   * @param opts
   */
  async cert(
    from: Account,
    to: Account,
    opts = {
      allowCreation: true,
      confirmBeforeCreation: true,
    }
  ): Promise<string> {
    if (!from || !to) throw new Error("Missing argument 'from' or 'to' !");

    // Check currency
    const currency = this.network.currency;
    if (!currency) throw new Error('ERROR.CHECK_NETWORK_CONNECTION');

    // Check issuer != recipient
    if (from.address === to.address) {
      throw new Error('ERROR.SELF_CERTIFICATION');
    }

    // Get issuer account
    const issuerAccount = await this.getByAddress(from.address, { withMembership: true });

    // Check enough credit
    const fee = currency.fees.cert;
    if (fee > issuerAccount.data.free) {
      throw new Error('ERROR.NOT_ENOUGH_CREDIT');
    }

    console.info(`${this._logPrefix}Certifying...\nfrom: ${from.address}\nto ${to.address}`);

    // Get pair, and unlock it
    const issuerPair = keyring.getPair(issuerAccount.address);
    if (issuerPair.isLocked) {
      console.debug(`${this._logPrefix}Unlocking address ${from.address} ...`);
      const isAuth = await this.auth();
      if (!isAuth) throw new Error('ERROR.AUTH_REQUIRED');
      issuerPair.unlock(this._password);
    }

    // Detect certification type
    let certType: CertType = null;
    if (isNil(to.meta?.status) || isNil(to.meta?.index)) {
      if (opts.allowCreation === false) throw new Error('ERROR.SEND_TX_FAILED');

      if (opts?.confirmBeforeCreation !== false) {
        // TODO ask user confirmation to create identity ?
      }

      console.debug(this._logPrefix + 'Target identity not exists: creating...');
      certType = CertType.IdentityCreation;
    } else {
      // check if target has identity index (identity already created)
      switch (to.meta.status) {
        case IdentityStatusEnum.Revoked:
          console.log('can not certify revoked identity');
          break;
        case IdentityStatusEnum.Unconfirmed:
          console.log('can not certify unconfirmed identity');
          break;
        case IdentityStatusEnum.Unvalidated: {
          // special case for last certification: request distance evaluation
          const toAccount = <Account>{ address: to.address, meta: to.meta, data: to.data, publicKey: to.publicKey };
          const toAccountRefereePercent = await this.wot.getRefereePercentByAccount(toAccount);
          if (toAccountRefereePercent > this.network.currency.params.minAccessibleReferees / 1000000000) {
            certType = CertType.IdentityCreation;
          } else {
            certType = CertType.CertCreation;
          }
          break;
        }
        case IdentityStatusEnum.Notmember: {
          // prevent certifying if lost membership for membership non renewal
          const toAccountRefereePercent = await this.wot.getRefereePercentByAccount(to);
          if (toAccountRefereePercent > this.network.currency.params.minAccessibleReferees / 1000000000) {
            certType = CertType.CertRenewal;
          } else {
            certType = CertType.CertCreation;
          }
          break;
        }
        default: {
          // check if certification already exists (renewal)
          const isRenewal = await this.isCertValid({ receiver: to.address, issuer: from.address });
          if (isRenewal) {
            certType = CertType.CertRenewal;
          } else {
            certType = CertType.CertCreation;
          }
          break;
        }
      }
    }

    await this.ready();

    try {
      let tx = null;
      switch (certType) {
        case CertType.IdentityCreation: {
          const txCertify = this.api.tx.identity.addCert(to.meta.index);
          const txIdentity = this.api.tx.identity.createIdentity(to.address);
          tx = this.api.tx.utility.batch([txCertify, txIdentity]);
          break;
        }
        case CertType.CertCreation:
          tx = this.api.tx.certification.addCert(to.meta.index);
          break;
        case CertType.CertRenewal:
          tx = this.api.tx.certification.renewCert(to.meta.index);
          break;
      }
      const status = (await ExtrinsicUtils.submit(tx, issuerPair)).status;
      console.info(`${this._logPrefix}Extrinsic status`, status.toHuman());
      console.info(`${this._logPrefix}Certifying completed at block hash #${status.hash.toHuman()}`);

      return status.hash.toHuman();
    } catch (err) {
      const error = new ExtrinsicError(this.api, err, 'ERROR.SEND_CERT_FAILED');
      console.error(`${this._logPrefix}Cannot certify: ${error?.message || error}`);
      throw error;
    }
  }

  /// WIP should wot action be part of account.service or wot.service?
  async confirm(account: Partial<Account>, uid: string): Promise<void> {
    // TODO some checks:
    // - status Unconfirmed
    // - uid availability

    // THIS IS DUPLICATED CODE
    const issuerPair = keyring.getPair(account.address);
    if (issuerPair.isLocked) {
      console.debug(`[account-service] Unlocking address ${account.address} ...`);
      const isAuth = await this.auth();
      if (!isAuth) throw new Error('ERROR.AUTH_REQUIRED');
      issuerPair.unlock(this._password);
    }

    // build tx
    const tx = this.api.tx.identity.confirmIdentity(uid);

    // try run tx (also code duplication)
    try {
      const { status } = await ExtrinsicUtils.submit(tx, issuerPair);
      console.info(`${this._logPrefix}Extrinsic status`, status.toHuman());
    } catch (err) {
      const error = new ExtrinsicError(this.api, err, 'ERROR.SEND_CERT_FAILED');
      console.error(`${this._logPrefix}Cannot confirm: ${error?.message || error}`);
      throw error;
    }

    // TODO process status
  }

  public async getBalance(addressOrAccount: string | Account): Promise<number> {
    if (!this.started) await this.ready();

    const account = typeof addressOrAccount === 'string' ? { address: addressOrAccount } : (addressOrAccount as Account);
    await this.loadData(account, { withBalance: true });
    return AccountUtils.getBalance(account);
  }

  async scanDerivations(mnemonic: string, numberOfDerivations = 30): Promise<DerivationAccount[]> {
    if (isNilOrBlank(mnemonic)) throw new Error("Missing argument 'mnemonic'");
    if (numberOfDerivations <= 0) throw new Error("Invalid argument 'numberOfDerivations'");
    const derivationAccounts: DerivationAccount[] = [];

    for (let i = -1; i <= numberOfDerivations; i++) {
      const derivationPath = i === -1 ? '' : `//${i}`;
      const address = this.generateAddress(`${mnemonic}${derivationPath}`);
      const shortAddress = formatAddress(address);
      derivationAccounts.push({ derivation: derivationPath, address, meta: { name: shortAddress } });
    }
    // Load balances data (e.g. balance)
    await Promise.all(derivationAccounts.map((account) => this.loadData(account, { withBalance: true })));

    // Filter to exclude balance === 0
    return derivationAccounts.filter((account) => AccountUtils.getBalance(account) > 0);
  }

  async isCertValid(filter: CertificationSearchFilter) {
    const certs = await firstValueFrom(this.indexer.certsSearch({ receiver: filter.receiver }));
    return isNotEmptyArray(certs.data.filter((cert) => cert.account.address === filter.issuer));
  }

  async claimUds(account: Account) {
    const issuerPair = keyring.getPair(account.address);
    if (issuerPair.isLocked) {
      console.debug(`[account-service] Unlocking address ${account.address} ...`);
      const isAuth = await this.auth();
      if (!isAuth) throw new Error('ERROR.AUTH_REQUIRED');
      issuerPair.unlock(this._password);
    }

    try {
      const tx = this.api.tx.universalDividend.claimUds();
      const { status } = await ExtrinsicUtils.submit(tx, issuerPair);
      console.info(`${this._logPrefix}Extrinsic status`, status.toHuman());
      return status.toHuman();
    } catch (err) {
      console.error(err);
      throw new Error('ERROR.CLAIM_UD_FAILED');
    }
  }

  async getBiometricAuth(): Promise<boolean> {
    try {
      await BiometricAuth.authenticate({
        reason: '',
        cancelTitle: this.translate.instant('COMMON.BTN_CANCEL'),
        allowDeviceCredential: true,
        iosFallbackTitle: '',
        androidTitle: this.translate.instant('ACCOUNT.SECURITY.BIOMETRIC.TITLE'),
        androidSubtitle: this.translate.instant('ACCOUNT.SECURITY.BIOMETRIC.LABEL'),
        androidConfirmationRequired: false,
      });
      return true;
    } catch (error) {
      throw new Error(error);
    }
  }

  async canBiometric(): Promise<boolean> {
    const info = await BiometricAuth.checkBiometry();
    return info.isAvailable;
  }

  async resetBiometricAuth() {
    this.settings.patchValue({ biometricAuth: false });
  }

  async isBiometricEnabled(): Promise<boolean> {
    return this.settings.get('biometricAuth') && isNotNilOrBlank(this.storage.get('validationCode'));
  }

  async getActionConfirmation(mobile: boolean, from: Account, retryUsed?: number): Promise<boolean> {
    let validation = false;
    const maxRetry = 3;
    retryUsed = retryUsed ?? 0;
    const isBiometricEnabled = await this.isBiometricEnabled();
    if (mobile && isBiometricEnabled) {
      validation = await this.getBiometricAuth();
    } else {
      const modal = await this.modalCtrl.create({
        component: ValidationModal,
      });
      await modal.present();

      const { data } = await modal.onWillDismiss();
      const codeValidation = data.codeValidation;

      if (isNotNilOrNaN(codeValidation)) {
        const code = await this.storage.get('validationCode');
        validation = codeValidation === code;
      }
    }

    if (validation === false && retryUsed < maxRetry) {
      retryUsed++;
      // if (mobile) {
      const { role } = await this.openRetryAlert(maxRetry - retryUsed);
      if (role === 'cancel') return false;
      // }

      validation = await this.getActionConfirmation(mobile, from, retryUsed);
    }

    if (!validation && isBiometricEnabled) {
      await this.resetBiometricAuth();
    }
    return validation;
  }

  private async openRetryAlert(retry: number) {
    const alert = await this.alertCtrl.create({
      header: this.translate.instant('ACCOUNT.SECURITY.WAITING_CONFIRMATION.TITLE'),
      message: this.translate.instant('ACCOUNT.SECURITY.WAITING_CONFIRMATION.ERROR', { retry: retry }),
      buttons: [
        {
          text: this.translate.instant('COMMON.BTN_NO'),
          role: 'cancel',
          handler: () => {
            return false;
          },
        },
        {
          text: this.translate.instant('COMMON.BTN_YES'),

          role: 'confirm',
          handler: () => {
            return true;
          },
        },
      ],
    });
    await alert.present();
    return await alert.onDidDismiss();
  }

  /**
   * Load account data (balance, tx history, etc.).
   * This load can be skipped, when data already loaded (See options)
   * @param account
   * @param opts
   * @private
   */
  async loadData(account: Account, opts?: LoadAccountDataOptions): Promise<Account> {
    if (!account?.address) throw new Error("Missing argument 'account' or 'account.address'");
    opts = {
      reload: false,
      withBalance: true,
      withTx: false, // disable by default
      withProfile: false,
      ...opts,
    };

    try {
      const now = Date.now();
      let changed = false;

      // Load balance (free + reserved)
      if (opts.withBalance === true && (isNil(account.data?.free) || opts.reload === true)) {
        console.debug(`${this._logPrefix} Loading ${account.meta?.name || formatAddress(account.address)} data...`);
        const { data } = await this.api.query.system.account(account.address);
        account.data = {
          ...account.data,
          ...data.toJSON(),
        };
        changed = true;

        //console.log('TODO', Object.keys(this.api.query));
        //await this.api.query.udAccountsStorage.udAccounts(account.address);
      }

      if (opts.withMembership === true && (isNil(account.meta.isMember) || opts.reload === true)) {
        const indexedAccount = await firstValueFrom(
          this.indexer
            .wotSearch({ address: account.address }, { first: 1, fetchPolicy: 'network-only' })
            .pipe(map(({ data }) => firstArrayValue(data)))
        );
        if (indexedAccount) {
          account.meta = {
            ...account.meta,
            uid: indexedAccount.meta?.uid,
            index: indexedAccount.meta?.index,
            isMember: indexedAccount.meta?.isMember,
            status: indexedAccount.meta?.status,
            createdOn: indexedAccount.meta?.createdOn,
          };
          changed = true;
        }
      }

      // Load TX
      if (opts.withTx === true && (isNil(account.data?.txs) || opts.reload === true)) {
        console.debug(`${this._logPrefix} Loading ${formatAddress(account.address)} TX history...`);
        console.warn('[account-service] TODO - Implement load TX history');

        // TODO
        //somethingLoaded = true;
      }

      // Load Cert
      if (opts.withCert === true) {
        // const certs = await this.api.query.cert.certsByReceiver(account.address);
        // console.debug(`${this._logPrefix} Loaded certs:`, certs);
      }

      // Load profile
      if (opts.withProfile === true && (isNil(account.meta.profile) || opts.reload === true)) {
        const { data } = await firstValueFrom(this.pod.profileSearch({ address: account.address }, { first: 1 }));
        const profile = firstArrayValue(data);
        if (profile?.meta) {
          account.meta = {
            ...account.meta,
            profile: {
              name: profile.meta.name,
              avatar: profile.meta.avatar,
            },
          };
          changed = true;
        } else {
          account.meta = {
            ...account.meta,
            profile: {}, // Remember that we already try to load it
          };
        }
      }

      // Emit change event
      if (changed && this.accounts) {
        console.debug(`${this._logPrefix} Loading ${formatAddress(account.address)} data [OK] in ${Date.now() - now}ms`, account);
        account.changes?.next();
      }
    } catch (err) {
      console.error(`${this._logPrefix}Failed to load ${formatAddress(account.address)} data:`, err);
      throw new Error('ERROR.LOAD_WALLET_DATA_ERROR');
    }

    return account;
  }

  private async refreshData(accounts: Account | Account[], opts?: LoadAccountDataOptions) {
    const array = Array.isArray(accounts) ? accounts : [accounts];
    try {
      const jobs = array.map((account) => this.loadData(account, opts));
      await Promise.all(jobs);

      // Notify accounts changed
      if (!opts || opts.emitEvent !== false) this.notifyChanged();
    } catch (err) {
      console.error(
        `${this._logPrefix}Failed to refresh data of:`,
        array.map((a) => formatAddress(a.address))
      );
      throw new Error('ERROR.UPDATE_WALLET_LIST_FAILED');
    }
  }

  private notifyChanged() {
    this.set('accounts', (s) => s.accounts.slice() /*create a copy*/);
  }

  forgetAll() {
    if (environment.production) {
      (this.accounts || []).forEach((account) => {
        keyring.forgetAccount(account.address);
      });
    }
    this.accounts = [];
  }
}
