import { ChangeDetectionStrategy, Component, Inject, OnInit } from '@angular/core';
import { FormBuilder, Validators } from '@angular/forms';
import { AppForm } from '@app/shared/form.class';
import { SettingsService } from '@app/settings/settings.service';
import { environment } from '@environments/environment';
import { FormUtils } from '@app/shared/forms';
import { isNil, isNotNilOrBlank } from '@app/shared/functions';
import { setTimeout } from '@rx-angular/cdk/zone-less/browser';
import { AuthData } from '@app/account/auth/auth.model';
import { SharedValidators } from '@app/shared/form/form-validators';
import { APP_WOT_CONTROLLER, IWotController } from '@app/wot/wot.model';
import { Clipboard } from '@capacitor/clipboard';

@Component({
  selector: 'app-address-form',
  templateUrl: 'address.form.html',
  changeDetection: ChangeDetectionStrategy.OnPush,
})
export class AddressForm extends AppForm<AuthData> implements OnInit {
  constructor(
    settings: SettingsService,
    formBuilder: FormBuilder,
    @Inject(APP_WOT_CONTROLLER) private wotCtrl: IWotController
  ) {
    super(
      formBuilder.group({
        address: [null, Validators.compose([Validators.required, SharedValidators.address])],
      })
    );

    this.mobile = settings.mobile;
    this._enable = true;
  }

  disable(opts?: { onlySelf?: boolean; emitEvent?: boolean }) {
    super.disable(opts);
  }

  ngOnInit() {
    super.ngOnInit();

    // For DEV only: set the default user, for testing
    if (!environment.production && environment.dev?.auth) {
      this.form.patchValue(environment.dev.auth);
    }
  }

  doCancel() {
    this.cancel.emit();
  }

  async doSubmit(event?: UIEvent) {
    if (event) {
      event.preventDefault();
      event.stopPropagation();
    }

    if (this.loading) {
      return;
    }

    if (!this.form.valid) {
      await FormUtils.waitWhilePending(this.form);
      if (this.form.invalid) {
        FormUtils.logErrors(this.form);
        return; // Skip if invalid
      }
    }

    this.markAsLoading();
    const data = this.value;
    this.error = null; // Reset error

    setTimeout(() => this.validate.emit(data));
  }

  get value(): AuthData {
    const data = this.form.value;
    return {
      v2: {
        mnemonic: data.mnemonic,
      },
    };
  }

  async pasteFromClipboard() {
    try {
      const { value } = await Clipboard.read();

      if (isNotNilOrBlank(value)) {
        this.form.patchValue({ address: value.trim() });
      }
    } catch (err) {
      console.error('Can not read clipboard', err);
    }
  }

  // get address corresponding to form input
  get address(): string {
    const data = this.form.value;
    // prevent displaying for empty credentials
    if (isNil(data.mnemonic)) {
      return '';
    }
    return '';
  }

  /* -- protected functions -- */

  async showWotModal(event: Event) {
    event.preventDefault();
    const searchText = this.form.get('address').value;

    setTimeout(async () => {
      console.info('[auth] Searching in wot, on ' + searchText);
      const data = await this.wotCtrl.select({ searchText, showItemActions: false, showFilterButtons: false, showSearchBar: true, autoLoad: true });

      if (!data) return; // User cancelled

      console.info('[auth] Selected account', data);
    }, 200);
  }

  protected markForCheck() {
    this._cd.markForCheck();
  }
}
