import { Injectable } from '@angular/core';
import { Peer, Peers } from '@app/shared/services/network/peer.model';
import { Promise } from '@rx-angular/cdk/zone-less/browser';
import { SettingsService } from '@app/settings/settings.service';
import { arrayRandomPick, isNotNil } from '@app/shared/functions';
import { StorageService } from '@app/shared/services/storage/storage.service';
import { Observable } from 'rxjs';
import { RxStateProperty, RxStateSelect } from '@app/shared/decorator/state.decorator';
import { RxStartableService } from '@app/shared/services/rx-startable-service.class';
import { HttpClient } from '@angular/common/http';

export interface IpfsState {
  peer: Peer;
  gatewayBaseUrl: string;
  offline: boolean;
}

@Injectable({ providedIn: 'root' })
export class IpfsService extends RxStartableService<IpfsState> {
  @RxStateSelect() peer$: Observable<Peer>;
  @RxStateProperty() peer: Peer;
  @RxStateProperty() gatewayBaseUrl: string;

  constructor(
    storage: StorageService,
    private settings: SettingsService,
    private httpClient: HttpClient
  ) {
    super(storage, {
      name: 'ipfs-service',
      startByReadyFunction: false, // Need an explicit call to start()
    });
  }

  getGatewayUrl(cid: string): string {
    return this.gatewayBaseUrl + cid;
  }

  /* -- protected functions -- */

  protected async ngOnStart(): Promise<IpfsState> {
    // Wait settings and storage
    const settings = await this.settings.ready();

    let peer = Peers.fromUri(settings.ipfsGateway);
    const isPeerAlive = await this.isPeerAlive(peer);

    if (!peer || !isPeerAlive) {
      const preferredFiltered = settings.preferredIpfsGateways.filter((preferred) => preferred !== settings.ipfsGateway);
      if (preferredFiltered.length === 0) throw new Error('ERROR.CHECK_NETWORK_CONNECTION');

      const peers = await this.filterAlivePeers(settings.preferredPods);
      if (!peers.length) {
        throw new Error('ERROR.CHECK_NETWORK_CONNECTION');
      }
      peer = arrayRandomPick(peers);
    }
    const gatewayBaseUrl = Peers.getHttpUri(peer) + '/ipfs/';

    return {
      peer,
      gatewayBaseUrl,
      offline: false,
    };
  }

  protected async filterAlivePeers(peers: string[]): Promise<Peer[]> {
    return (
      await Promise.all(peers.map((peer) => Peers.fromUri(peer)).map((peer) => this.isPeerAlive(peer).then((alive) => (alive ? peer : undefined))))
    ).filter(isNotNil);
  }

  protected async isPeerAlive(peer: Peer): Promise<boolean> {
    const url = Peers.getHttpUri(peer) + '/api/v0/version';

    try {
      await this.httpClient.get(url, {}).toPromise();
      return true;
    } catch (err) {
      console.error(`${this._logPrefix}Le pair IPFS n'est pas accessible:`, peer, err);
      return false;
    }
  }
}
