import { inject, Injectable } from '@angular/core';
import { ApiPromise, WsProvider } from '@polkadot/api';
import { SettingsService } from '../settings/settings.service';
import { Peer, Peers } from '@app/shared/services/network/peer.model';
import { abbreviate, WELL_KNOWN_CURRENCIES } from '@app/shared/currencies';
import { Currency } from '../currency/currency.model';
import { RxStartableService } from '@app/shared/services/rx-startable-service.class';
import { RxStateProperty, RxStateSelect } from '@app/shared/decorator/state.decorator';
import { EMPTY, firstValueFrom, mergeMap, Observable, race, tap } from 'rxjs';
import { catchError, filter, map } from 'rxjs/operators';
import { arrayRandomPick, isEmptyArray, isNil, isNotNil, isNotNilOrBlank, toNumber } from '@app/shared/functions';
import { IndexerService } from './indexer/indexer.service';
import { fromDateISOString } from '@app/shared/dates';
import { ContextService } from '@app/shared/services/storage/context.service';
import { PodService } from '@app/network/pod/pod.service';
import { HttpClient } from '@angular/common/http';
import { environment } from '@environments/environment';
import moment from 'moment';
import DiscourseAPI from 'discourse2';
import { JsonFeedItem, JsonTopic, Topic } from '@app/home/feed/feed.model';

import { Promise, setTimeout } from '@rx-angular/cdk/zone-less/browser';

export interface NetworkState {
  peer: Peer;
  currency: Currency;
  currencySymbol: string;
  currentUd?: number;
  api: ApiPromise;
}

export interface DistancePrecomputionData {
  height: number;
  referees_count: number;
  min_certs_for_referee: number;
  results: Map<number, number>;
}

@Injectable({ providedIn: 'root' })
export class NetworkService extends RxStartableService<NetworkState> {
  indexer = inject(IndexerService);
  pod = inject(PodService);

  @RxStateProperty() peer: Peer;
  @RxStateProperty() currency: Currency;
  @RxStateProperty() currencySymbol: string;
  @RxStateProperty() currentUd: number;
  @RxStateProperty() api: ApiPromise;

  @RxStateSelect() peer$: Observable<Peer>;
  @RxStateSelect() currency$: Observable<Currency>;
  @RxStateSelect() currentUd$: Observable<number>;

  constructor(
    private settings: SettingsService,
    private context: ContextService,
    private http: HttpClient
  ) {
    super(settings, {
      name: 'network-service',
    });

    this.connect('currencySymbol', this.currency$.pipe(map((currency) => currency?.symbol)));

    // Restart when settings peer changed
    this.hold(
      this.settings.peer$.pipe(
        filter((peer) => isNotNilOrBlank(peer) && this.started),
        map(Peers.fromUri),
        filter((peer) => !Peers.equals(this.peer, peer))
      ),
      () => this.restart()
    );

    this.hold(this.currency$, (currency) => (this.context.currency = currency));

    this.connect(
      'currentUd',
      this.select('currency').pipe(
        mergeMap(async (currency) => {
          const ud = await this.api.query.universalDividend.currentUd();
          return toNumber(ud) / currency.powBase;
        }),
        tap((currentUd) => console.info(`${this._logPrefix}Current UD: ${currentUd}`))
      )
    );
  }

  protected async ngOnStart(): Promise<NetworkState> {
    const settings = await this.settings.ready();

    let isPeerAlive: boolean;
    let peer = Peers.fromUri(settings.peer);

    if (isNil(peer)) {
      isPeerAlive = false;
    } else {
      isPeerAlive = await this.isPeerAlive(peer);
    }

    if (!peer || !isPeerAlive) {
      const preferredFiltered = settings.preferredPeers.filter((preferred) => preferred !== settings.peer);
      if (preferredFiltered.length === 0) throw new Error('ERROR.CHECK_NETWORK_CONNECTION');

      const peers = await this.filterAlivePeers(settings.preferredPeers);
      if (!peers.length) {
        throw new Error('ERROR.CHECK_NETWORK_CONNECTION');
      }
      peer = arrayRandomPick(peers);
    }
    const wsUri = Peers.getWsUri(peer);
    console.info(`${this._logPrefix}Connecting to peer {${wsUri}}...`);

    settings.peer = wsUri;

    // Extract all types from definitions - fast and dirty approach, flatted on 'types'
    // const types = Object.values(definitions).reduce((res: any, { types }): object => {
    //   return { ...res, ...types };
    // }, {});
    // this.log(types);

    // Construct
    const wsProvider = new WsProvider(wsUri);
    const api = await ApiPromise.create({
      provider: wsProvider,
      //,...types
    });

    // Log API
    //console.debug(`${this._logPrefix}API loaded [${Object.keys(api).join(',')}]`)

    // Get the chain information
    const [chain, nodeName, nodeVersion, properties] = await Promise.all([
      api.rpc.system.chain(),
      api.rpc.system.name(),
      api.rpc.system.version(),
      api.rpc.system.properties(),
    ]);
    //const chainObj = await api.rpc.system.chain();
    //const chainPrefix = '' + chain.toHuman().split(' ')?.[0];
    const genesis = api.genesisHash.toHex();

    console.info(`${this._logPrefix}Node {${nodeName}} v${nodeVersion}`);
    console.info(`${this._logPrefix}Connecting to chain {${chain}}: ` + JSON.stringify(properties.toHuman()));

    let currency: Currency;
    // Check is well known currency
    const wellKnownCurrency = Object.values(WELL_KNOWN_CURRENCIES).find((c) => c.displayName === chain.toHuman());
    if (wellKnownCurrency) {
      if (wellKnownCurrency.genesis && wellKnownCurrency.genesis !== genesis) {
        console.warn(`${this._logPrefix}Invalid genesis for ${chain}! Expected ${wellKnownCurrency.genesis} but peer return ${genesis}`);
      }
      currency = <Currency>{ ...wellKnownCurrency };
    } else {
      console.warn(`${this._logPrefix}Not a well known currency: ${chain}!`);
    }
    currency = currency || <Currency>{};
    currency.displayName = currency?.displayName || chain.toHuman();
    currency.symbol = currency?.symbol || properties.tokenSymbol.value?.[0].toHuman() || abbreviate(this.currency.displayName);
    currency.decimals = currency?.decimals || +properties.tokenDecimals.value?.[0].toHuman() || 0;
    currency.powBase = Math.pow(10, currency.decimals);
    currency.ss58Format = currency.ss58Format || WELL_KNOWN_CURRENCIES.GDEV.ss58Format; // TODO use G1 defaults
    currency.genesis = genesis;
    currency.startTime = fromDateISOString(currency.startTime);
    currency.fees = {
      ...WELL_KNOWN_CURRENCIES.GDEV.fees, // TODO use G1 defaults
      ...(currency.fees || {}),
    };
    currency.params = {
      minAccessibleReferees: toNumber(api.consts.distance.minAccessibleReferees),
    };
    // Read the genesys block hash
    console.debug(`${this._logPrefix}Chain genesis: ${currency.genesis}`);

    // Retrieve the latest header
    const lastHeader = await api.rpc.chain.getHeader();
    console.info(`${this._logPrefix}Last block: #${lastHeader.number} - hash ${lastHeader.hash}`);

    const ud0 = toNumber(api.consts.universalDividend.unitsPerUd) / currency.powBase;

    // Configure and start indexer and pod
    this.indexer.currency = currency;
    this.indexer.api = api;
    this.pod.currency = currency;
    await Promise.all([
      this.indexer.start(),
      // this.pod.start()
    ]).catch((err) => console.error(`${this._logPrefix}Error starting indexer or pod:`, err));

    return {
      api,
      peer,
      currency,
      currencySymbol: currency?.symbol,
      currentUd: ud0,
    };
  }

  protected async ngOnStop(): Promise<void> {
    await this.indexer.stop();

    return super.ngOnStop();
  }

  protected async filterAlivePeers(peers: string[]): Promise<Peer[]> {
    return (
      await Promise.all(peers.map((peer) => Peers.fromUri(peer)).map((peer) => this.isPeerAlive(peer).then((alive) => (alive ? peer : undefined))))
    ).filter(isNotNil);
  }

  protected async isPeerAlive(peer: Peer): Promise<boolean> {
    const timeoutDuration = 4000;

    return new Promise<boolean>((resolve) => {
      const timeoutId = setTimeout(() => {
        ws.close();
        resolve(false);
      }, timeoutDuration);

      const wsUri = Peers.getWsUri(peer);
      const ws = new WebSocket(wsUri);

      ws.onopen = () => {
        const healthRequest = {
          id: 1,
          jsonrpc: '2.0',
          method: 'system_health',
          params: [],
        };
        ws.send(JSON.stringify(healthRequest));
      };

      ws.onmessage = () => {
        clearTimeout(timeoutId);
        ws.close();
        resolve(true);
      };

      ws.onerror = () => {
        clearTimeout(timeoutId);
        ws.close();
        resolve(false);
      };
    });
  }

  async race<T>(urls: string[]): Promise<T> {
    try {
      const requests = (urls || []).map((url) => this.http.get<T>(url).pipe(catchError(() => EMPTY)));
      return await firstValueFrom(race<T[]>(...requests));
    } catch (err) {
      return null;
    }
  }

  async getDistancePrecomputionData(): Promise<DistancePrecomputionData> {
    const distanceFileUrls = this.settings.data.preferredDistanceFileUrls;
    if (isEmptyArray(distanceFileUrls)) {
      console.warn(`${this._logPrefix}No precomputed distance file URLs found, in settings and environment!`);
      return Promise.resolve(null);
    }
    console.debug(`${this._logPrefix}Fetching precomputed distance data...`, distanceFileUrls);
    return this.race<DistancePrecomputionData>(distanceFileUrls);
  }

  async loadJsonFeedUrls(urls: string[]): Promise<JsonTopic[]> {
    const topicsPromises = (urls || []).map(async (url) => {
      let topics: JsonTopic[] = [];
      try {
        if (this.isDiscourseCategory(url)) {
          console.log(`${this._logPrefix}Une catégorie a été détectée dans l'url: ${url}`);
          topics = await this.getTopicsByCategory(url);
        } else {
          topics = await this.getPostByUrl(url);
        }
        return topics;
      } catch (err) {
        console.error(`${this._logPrefix}`, err);
      }
    });
    const allTopics = await Promise.all(topicsPromises);
    // Aplatir le tableau en 1 dimension, car chaque URL retourne un tableau de topics
    return allTopics.flat();
  }

  protected async getPostByUrl(url: string): Promise<JsonTopic[]> {
    const maxAgeInMonths = environment.feed.maxAgeInMonths;
    const minDate = maxAgeInMonths > 0 ? moment().subtract(maxAgeInMonths, 'month').startOf('day').utc() : undefined;
    const topicId = this.extractTopicIdFromUrl(url);
    if (!topicId) {
      console.error(`${this._logPrefix}Invalid URL`);
      return undefined;
    }
    const discourse = new DiscourseAPI(environment.forumUrl);
    const topic = await discourse.getTopic({ id: topicId.toString() });
    const posts = topic.post_stream.posts;
    if (posts.length === 0) {
      console.error(`${this._logPrefix}Aucun post trouvé dans ce topic.`);
      return undefined;
    }
    const lastPost = posts[posts.length - 1];
    if (moment(lastPost.created_at).isBefore(minDate)) {
      console.log(`${this._logPrefix}Le post ne réponds pas la régle de date`);
      return undefined;
    }
    return [this.formatTopics(lastPost, topic, url)];
  }

  private async getTopicsByCategory(url: string) {
    console.log(`${this._logPrefix}Une catégorie a été détectée dans l'url: ${url}`);

    const categoryId = this.extractCategoryIdFromUrl(url);
    const categorySlug = this.extractCategorySlugFromUrl(url);
    const discourse = new DiscourseAPI(environment.forumUrl);
    const category = await discourse.listCategoryTopics({ id: categoryId, slug: categorySlug });

    const topics = category.topic_list.topics.filter(
      (topic: Topic) =>
        topic.pinned &&
        topic.visible &&
        !topic.tags.includes('about-category') &&
        this.getLanguageFromTitle(topic.fancy_title) != this.settings.locale
    );
    return Promise.all(
      topics.map(async (t) => {
        const topic = await discourse.getTopic({ id: t.id.toString() });
        const posts = topic.post_stream.posts;
        if (posts.length === 0) {
          console.error(`${this._logPrefix}Aucun post trouvé dans ce topic.`);
          return undefined;
        }
        return this.formatTopics(posts[0], topic, url);
      })
    );
  }

  protected formatTopics(post, topic, url: string) {
    const feedItem = this.prepareFeedItem(post, topic);
    return <JsonTopic>{
      fancyTitle: this.cleanTitle(topic.fancy_title),
      archived: topic.archived,
      author: topic.details.created_by.name,
      authorAvatar: environment.forumUrl + topic.details.created_by.avatar_template.replace('{size}', '48'),
      tags: topic.tags,
      createdAt: moment(topic.created_at),
      updatedAt: moment(topic.last_posted_at),
      feedItem: feedItem,
      feedUrl: url,
    };
  }

  protected isDiscourseCategory(url: string) {
    return url.match(/\/c\//);
  }

  protected prepareFeedItem(post, topic): JsonFeedItem {
    const maxContentLength = environment.feed.maxContentLength;
    if (maxContentLength > 0 && post.cooked && post.cooked.length > maxContentLength) {
      const endIndex = Math.max(post.cooked.lastIndexOf(' ', maxContentLength), post.cooked.lastIndexOf('<', maxContentLength));
      post.cooked = post.cooked.substring(0, endIndex) + ' (...)';
      post.truncated = true;
    }
    post.name = post.name ?? topic.details.last_poster.name;
    return <JsonFeedItem>{
      author: post.name,
      authorAvatar:
        environment.forumUrl + (post.avatar_template.replace('{size}', '48') ?? topic.details.last_poster.avatar_template.replace('{size}', '48')),
      content_html: post.cooked,
      date_published: post.created_at,
      truncated: post.truncated,
    };
  }

  protected extractCategoryIdFromUrl(url: string) {
    const idMatch = url.match(/\/(\d+)$/);
    return idMatch ? parseInt(idMatch[1], 10) : null;
  }

  protected extractCategorySlugFromUrl(url: string) {
    const match = url.match(/\/c\/([\w/-]+)\/\d+$/);
    return match ? match[1] : null;
  }

  protected extractTopicIdFromUrl(url: string) {
    const idMatch = url.match(/\/(\d+)(?:$|\/)/) || url.match(/\/(\d+)\.json$/);
    return idMatch ? parseInt(idMatch[1], 10) : null;
  }

  protected getLanguageFromTitle(title: string) {
    if (!title) return '';

    const matches = title.match('\\s\\(([a-z]{2}(?:-[A-Z]{2})?)\\)$');
    return matches && matches[1];
  }

  protected cleanTitle(title: string) {
    if (!title) return undefined;
    return title.replace(/\s\([a-z]{2}(:?-[A-Z]{2})?\)$/, '');
  }
}
