import { Component, Inject, Input, OnInit, ViewChild } from '@angular/core';
import { APP_LOCALES, LocaleConfig, PeriodOption, Settings } from '@app/settings/settings.model';
import { AppPage, AppPageState } from '@app/shared/pages/base-page.class';
import { RxState } from '@rx-angular/state';
import { RxStateProperty, RxStateSelect } from '@app/shared/decorator/state.decorator';
import { Observable, skip } from 'rxjs';
import { IonModal, ModalController } from '@ionic/angular';
import { NetworkService } from '@app/network/network.service';
import { map } from 'rxjs/operators';
import moment from 'moment';
import { AccountsService } from '@app/account/accounts.service';
import { isNotNil } from '@app/shared/functions';
import { Account } from '@app/account/account.model';
import DurationConstructor = moment.unitOfTime.DurationConstructor;

export interface SettingsPageState extends Settings, AppPageState {
  useRelativeUnit: boolean;
  dirty: boolean;
  newPeer: string;
}

@Component({
  selector: 'app-settings',
  templateUrl: './settings.page.html',
  styleUrls: ['./settings.page.scss'],
  providers: [RxState],
})
export class SettingsPage extends AppPage<SettingsPageState> implements OnInit {
  unauthOptions = [
    {
      label: 'SETTINGS.KEEP_AUTH_OPTION.SECONDS',
      labelParam: 10,
      value: 10_000,
    },
    {
      label: 'SETTINGS.KEEP_AUTH_OPTION.SECONDS',
      labelParam: 30,
      value: 30_000,
    },
    {
      label: 'SETTINGS.KEEP_AUTH_OPTION.MINUTE',
      labelParam: 1,
      value: 60_000,
    },
    {
      label: 'SETTINGS.KEEP_AUTH_OPTION.MINUTES',
      labelParam: 15,
      value: 15 * 60_000,
    },
    {
      label: 'SETTINGS.KEEP_AUTH_OPTION.MINUTES',
      labelParam: 30,
      value: 30 * 60_000,
    },
    {
      label: 'SETTINGS.KEEP_AUTH_OPTION.ALWAYS',
      value: 9999 * 60_000,
    },
  ];

  redeemPeriodOptions: PeriodOption[] = [
    {
      key: 'day',
      i18nKey: 'SETTINGS.CLAIM_UD_PERIOD_OPTION.DAILY',
    },
    {
      key: 'week',
      i18nKey: 'SETTINGS.CLAIM_UD_PERIOD_OPTION.WEEKLY',
    },
    {
      key: 'month',
      i18nKey: 'SETTINGS.CLAIM_UD_PERIOD_OPTION.MONTHLY',
    },
  ];

  @RxStateSelect() preferredPeers$: Observable<string[]>;
  @RxStateSelect() peer$: Observable<string>;
  @RxStateSelect() preferredIndexers$: Observable<string[]>;
  @RxStateSelect() indexer$: Observable<string>;
  @RxStateSelect() preferredPods$: Observable<string[]>;
  @RxStateSelect() pod$: Observable<string>;
  @RxStateSelect() preferredIpfsGateways$: Observable<string[]>;
  @RxStateSelect() ipfsGateway$: Observable<string>;
  @RxStateSelect() dirty$: Observable<boolean>;
  @RxStateSelect() newPeer$: Observable<string>;
  @RxStateSelect() durationRedeemUd$: Observable<number>;
  @RxStateSelect() biometricAuth$: Observable<boolean>;

  @RxStateProperty() darkMode: boolean;
  @RxStateProperty() locale: string;
  @RxStateProperty() useRelativeUnit: boolean;
  @RxStateProperty() peer: string;
  @RxStateProperty() indexer: string;
  @RxStateProperty() pod: string;
  @RxStateProperty() ipfsGateway: string;
  @RxStateProperty() unAuthDelayMs: number;
  @RxStateProperty() dirty: boolean;
  @RxStateProperty() expertMode: boolean;
  @RxStateProperty() biometricAuth: boolean;
  @RxStateProperty() displayUnit: 'base' | 'du';
  @Input() @RxStateProperty() customPeer: string;
  @Input() @RxStateProperty() selectedRedeemPeriod: DurationConstructor = 'day';

  @ViewChild('selectPeerModal') selectPeerModal: IonModal;
  @ViewChild('selectIndexerModal') selectIndexerModal: IonModal;
  @ViewChild('selectPodModal') selectPodModal: IonModal;
  @ViewChild('selectIpfsGatewayModal') selectIpfsGatewayModal: IonModal;

  canBiometric = false;
  account: Account;
  constructor(
    protected networkService: NetworkService,
    protected accountService: AccountsService,
    protected modalCtrl: ModalController,
    @Inject(APP_LOCALES) protected locales: LocaleConfig[]
  ) {
    super({ name: 'settings' });

    // Conversion displayUnit <--> useRelativeUnit
    this._state.connect('displayUnit', this._state.select('useRelativeUnit').pipe(map((useRelativeUnit) => (useRelativeUnit ? 'du' : 'base'))));
    this._state.connect('useRelativeUnit', this._state.select('displayUnit').pipe(map((unit) => unit === 'du')));

    // Detect changes
    this._state.hold(
      this._state
        .select(['locale', 'peer', 'indexer', 'pod', 'ipfsGateway', 'unAuthDelayMs', 'displayUnit', 'selectedRedeemPeriod', 'expertMode'], (s) => s)
        .pipe(skip(1)),
      () => {
        if (this.mobile) {
          this.save();
        } else {
          this.markAsDirty();
        }
      }
    );
  }

  protected async ngOnLoad() {
    await this.settings.ready();
    await this.accountService.ready();
    this.canBiometric = (await this.accountService.canBiometric()) && this.accountService.isLogin;
    return {
      ...this.settings.clone(),
      dirty: false,
    };
  }

  cancel() {
    this._state.set(this.settings.clone());
    this.markForCheck();
  }

  save() {
    this.settings.patchValue(this.data);
    this.dirty = false;
  }

  setStateValue(value: string, property: keyof Settings) {
    this._state.set(property, () => value);
    return true;
  }

  onPeerSubmit(property: keyof Settings) {
    return this.setStateValue(this.customPeer, property);
  }

  markAsDirty() {
    this.dirty = true;
  }

  toggleDarkMode() {
    this.darkMode = !this.darkMode;
    this.save();
  }

  async toggleBiometricAuth() {
    this.account = await this.accountService.getDefault();
    if (this.biometricAuth) {
      const confirmation = await this.accountService.getActionConfirmation(this.mobile, this.account);
      if (confirmation) {
        await this.accountService.resetBiometricAuth();
      }
      return;
    }

    let validation = await this.accountService.getActionConfirmation(this.mobile, this.account);

    if (validation) {
      validation = await this.accountService.getBiometricAuth();
    }
    this.settings.patchValue({ biometricAuth: validation });
  }

  protected readonly isNotNil = isNotNil;
}
