import { ChangeDetectionStrategy, Component, Inject, Input, OnInit } from '@angular/core';

import { AppPage, AppPageState } from '@app/shared/pages/base-page.class';
import { Account } from '@app/account/account.model';
import { AccountsService } from '@app/account/accounts.service';
import { Clipboard } from '@capacitor/clipboard';
import { RxStateProperty, RxStateSelect } from '@app/shared/decorator/state.decorator';
import { firstValueFrom, mergeMap, Observable, switchMap } from 'rxjs';
import { RxState } from '@rx-angular/state';
import { APP_TRANSFER_CONTROLLER, ITransferController } from '@app/transfer/transfer.model';
import { filter, map } from 'rxjs/operators';
import { firstArrayValue, isNil, isNotEmptyArray, isNotNilOrBlank } from '@app/shared/functions';
import { IndexerService } from '@app/network/indexer/indexer.service';
import { address2PubkeyV1, pubkeyV1Checksum } from '@app/shared/currencies';
import { IdentityStatusEnum } from '@app/network/indexer/indexer-types.generated';
import { DistanceInformation, WotService } from '@app/wot/wot.service';
import { NetworkService } from '@app/network/network.service';
import { HelpPage } from '@app/home/help/help.page';
import { AlertController, ModalController } from '@ionic/angular';
import { VerificationModal } from '@app/certification/verification/verification.modal';

export interface WotDetailsPageState extends AppPageState {
  address: string;
  account: Account;
  receivedCertCount: number;
  givenCertCount: number;
  refereePercent: number;
  distanceInformation: DistanceInformation;
  accounts: Account[];
  isRenewal: boolean;
  canCertify: boolean;
}

@Component({
  selector: 'app-wot-details',
  templateUrl: './wot-details.page.html',
  styleUrls: ['./wot-details.page.scss'],
  changeDetection: ChangeDetectionStrategy.OnPush,
  providers: [RxState],
})
export class WotDetailsPage extends AppPage<WotDetailsPageState> implements OnInit {
  @RxStateSelect() address$: Observable<string>;
  @RxStateSelect() account$: Observable<Account>;
  @RxStateSelect() receivedCertCount$: Observable<number>;
  @RxStateSelect() givenCertCount$: Observable<number>;
  @RxStateSelect() refereePercent$: Observable<number>;
  @RxStateSelect() accounts$: Observable<Account[]>;
  @RxStateSelect() isRenewal$: Observable<boolean>;
  @RxStateSelect() canCertify$: Observable<boolean>;

  @Input() showToolbar = true;
  @Input() showBalance = false;
  @Input() showToastOnCertify = true;
  @Input() @RxStateProperty() address: string;
  @Input() @RxStateProperty() account: Account;
  @RxStateProperty() refereePercent: number;
  @RxStateProperty() distanceInformation: DistanceInformation;
  @RxStateProperty() accounts: Account[];
  @RxStateProperty() isRenewal: boolean;
  @RxStateProperty() canCertify: boolean;

  constructor(
    private accountsService: AccountsService,
    private indexerService: IndexerService,
    private networkService: NetworkService,
    private alertCtrl: AlertController,
    private wotService: WotService,
    private modalCtrl: ModalController,
    @Inject(APP_TRANSFER_CONTROLLER) private transferController: ITransferController
  ) {
    super({ name: 'wot-details-page' });
    this._state.connect('address', this.activatedRoute.paramMap.pipe(map((paramMap) => paramMap.get('address'))));

    this._state.connect(
      'account',
      this.address$.pipe(
        mergeMap(async (address) => {
          const ownedAddress = await this.accountsService.isAvailable(address);
          return { address, ownedAddress };
        }),
        switchMap(({ address, ownedAddress }) => {
          if (ownedAddress) {
            return this.accountsService.watchByAddress(address);
          }
          return this.indexerService.wotSearch({ address }, { first: 1 }).pipe(map(({ data }) => firstArrayValue(data)));
        }),
        mergeMap(async (account) => {
          if (account.data) return account;
          const { data } = await this.accountsService.api.query.system.account(account.address);
          return {
            ...account,
            data: {
              ...JSON.parse(data.toString()),
            },
          };
        })
      )
    );

    const validAddress$ = this.account$.pipe(
      map((account) => account?.address),
      filter(isNotNilOrBlank)
    );

    // Watch certification count
    this._state.connect(
      'receivedCertCount',
      validAddress$.pipe(
        switchMap((address) => this.indexerService.certsSearch({ receiver: address }, { first: 0 })),
        map(({ total }) => total)
      )
    );
    this._state.connect(
      'givenCertCount',
      validAddress$.pipe(
        switchMap((address) => this.indexerService.certsSearch({ issuer: address }, { first: 0 })),
        map(({ total }) => total)
      )
    );

    this._state.connect('accounts', this.accountsService.watchAll());
  }

  ngOnInit() {
    super.ngOnInit();
  }

  protected async ngOnLoad(): Promise<WotDetailsPageState> {
    const account = await firstValueFrom(this.account$);
    const accounts = await firstValueFrom(this.accounts$);

    const refereePercent = await this.wotService.getRefereePercentByAccount(account);
    const distanceInformation = await this.wotService.getDistanceInformation(account);
    const canCertify = await this.enableCertify();
    const isRenewal = await this.isRenewalCertification(accounts);
    return <WotDetailsPageState>{ account, accounts, refereePercent, distanceInformation, canCertify, isRenewal };
  }

  async copyPubkey(event: UIEvent, pubkey?: string) {
    if (this.loading) return;
    event.preventDefault();

    pubkey = pubkey || address2PubkeyV1(this.data?.address, this.context.currency?.ss58Format, false);
    if (!pubkey) return; // Skip

    // Append checksum
    pubkey += ':' + pubkeyV1Checksum(pubkey);

    await Clipboard.write({
      string: pubkey,
    });
    await this.showToast({ message: 'INFO.COPY_TO_CLIPBOARD_DONE' });
  }

  async copyAddress(event: UIEvent) {
    if (this.loading || !this.data?.account?.address) return; // Skip

    event.preventDefault();

    await Clipboard.write({
      string: this.account.address,
    });
    await this.showToast({ message: 'INFO.COPY_TO_CLIPBOARD_DONE' });
  }

  async transferTo() {
    return this.transferController.transfer({ recipient: this.account });
  }

  async certifyTo() {
    const accountsMember = this.accounts.filter((account) => account.meta.isMember === true);
    const issuer = accountsMember.length === 1 ? accountsMember[0] : await this.accountsService.selectAccount({ isMember: true });
    if (!issuer) {
      await this.showToast({
        id: 'cert',
        message: 'INFO.CERTIFICATION_PENDING',
        duration: -1,
      });
      return; // Skip}
    }
    const verification = await this.showModalVerification();
    if (!verification) return; // Skip

    this.markAsLoading();
    this.resetError();

    try {
      if (this.showToastOnCertify)
        await this.showToast({
          id: 'cert',
          message: 'INFO.CERTIFICATION_PENDING',
          duration: -1,
        });
      const certHash = await this.accountsService.cert(issuer, this.account);

      if (this.showToastOnCertify)
        await this.showToast({
          id: 'cert',
          message: 'INFO.CERTIFICATION_DONE',
          swipeGesture: 'vertical',
          color: 'secondary',
        });

      return certHash;
    } catch (err) {
      this.showErrorToast(err, { id: 'cert' });
      this.markAsLoaded();
    }
  }

  distanceRuleValid(): boolean {
    return this.disableDistanceRule() || this.refereePercent > this.networkService.currency.params.minAccessibleReferees / 1000000000;
  }

  disableDistanceRule(): boolean {
    return isNil(this.refereePercent);
  }

  titleDistanceRule() {
    return this.translate.instant('ACCOUNT.TITLE_DISTANCE_RULES', {
      refereesReached: this.distanceInformation?.refereesReached,
      refereesCount: this.distanceInformation?.refereesCount,
      height: this.distanceInformation?.height,
    });
  }

  async showModalDistanceRuleDefinition() {
    const modal = await this.modalCtrl.create({
      id: 'help-modal',
      component: HelpPage,
      componentProps: <HelpPage>{
        highlightedDefinition: 'GLOSSARY_DISTANCE_RULE',
      },
      presentingElement: this._presentingElement,
      canDismiss: true,
    });
    await modal.present();
  }

  private async showModalVerification(): Promise<boolean> {
    const modal = await this.modalCtrl.create({
      id: 'help-modal',
      component: VerificationModal,
      presentingElement: this._presentingElement,
      canDismiss: true,
    });
    await modal.present();

    const { data } = await modal.onDidDismiss();

    if (data === undefined) {
      return false;
    }
    if (data === false) {
      const warning = await this.alertCtrl.create({
        header: this.translate.instant('ACCOUNT.CERTIFICATION_MODAL.REMINDER_TITLE'),
        message: this.translate.instant('ACCOUNT.CERTIFICATION_MODAL.CHECKLIST_CONDITIONS_NOT_MET'),
        buttons: [
          {
            text: this.translate.instant('COMMON.BTN_OK'),
            role: 'confirm',
            cssClass: 'primary',
          },
        ],
      });
      await warning.present();
      return false;
    }

    return true;
  }

  protected async enableCertify() {
    const accounts = this.accounts.filter((account) => account.meta.isMember === true);
    return !isNotEmptyArray(accounts);
  }

  private async isRenewalCertification(accounts: Account[]) {
    return (
      this.canCertify &&
      (await this.accountsService.isCertValid({
        receiver: this.account?.address,
        issuer: accounts[0]?.address,
      }))
    );
  }

  protected readonly IdentityStatusEnum = IdentityStatusEnum;
}
